import { type GossipSub, type GossipsubMessage, gossipsub } from "@chainsafe/libp2p-gossipsub";
import {
	type TopicScoreParams,
	createPeerScoreParams,
	createTopicScoreParams,
} from "@chainsafe/libp2p-gossipsub/score";
import { noise } from "@chainsafe/libp2p-noise";
import { yamux } from "@chainsafe/libp2p-yamux";
import { autoTLS } from "@libp2p/auto-tls";
import { autoNAT } from "@libp2p/autonat";
import { type BootstrapComponents, bootstrap } from "@libp2p/bootstrap";
import { circuitRelayServer, circuitRelayTransport } from "@libp2p/circuit-relay-v2";
import { generateKeyPairFromSeed } from "@libp2p/crypto/keys";
import { dcutr } from "@libp2p/dcutr";
import { devToolsMetrics } from "@libp2p/devtools-metrics";
import { identify, identifyPush } from "@libp2p/identify";
import type {
	Address,
	EventCallback,
	PeerDiscovery,
	PeerId,
	Stream,
	StreamHandler,
	SubscriptionChangeData,
} from "@libp2p/interface";
import { keychain } from "@libp2p/keychain";
import { peerIdFromString } from "@libp2p/peer-id";
import { ping } from "@libp2p/ping";
import {
	type PubSubPeerDiscoveryComponents,
	pubsubPeerDiscovery,
} from "@libp2p/pubsub-peer-discovery";
import { webRTC, webRTCDirect } from "@libp2p/webrtc";
import { webSockets } from "@libp2p/websockets";
import * as filters from "@libp2p/websockets/filters";
import { webTransport } from "@libp2p/webtransport";
import { type MultiaddrInput, multiaddr } from "@multiformats/multiaddr";
import { WebRTC } from "@multiformats/multiaddr-matcher";
import { Logger, type LoggerOptions } from "@sfroment/logger";
import { type Libp2p, ServiceFactoryMap, createLibp2p } from "libp2p";
import { fromString as uint8ArrayFromString } from "uint8arrays/from-string";

import { Message } from "./proto/drp/network/v1/messages_pb.js";
import { uint8ArrayToStream } from "./stream.js";
import { waitForEvent } from "./utils/helper.js";

export * from "./stream.js";

export const DRP_MESSAGE_PROTOCOL = "/drp/message/0.0.1";
export const BOOTSTRAP_NODES = [
	"/dns4/bootstrap1.topology.gg/tcp/443/wss/p2p/12D3KooWBu1pZ3v2u6tXSmkN35kiMLENpv3bEXcyT1GJTVhipAkG",
	"/dns4/bootstrap2.topology.gg/tcp/443/wss/p2p/12D3KooWLGuTtCHLpd1SBHeyvzT3kHVe2dw8P7UdoXsfQHu8qvkf",
];
let log: Logger;

// snake_casing to match the JSON config
export interface DRPNetworkNodeConfig {
	announce_addresses?: string[];
	bootstrap?: boolean;
	bootstrap_peers?: string[];
	browser_metrics?: boolean;
	listen_addresses?: string[];
	log_config?: LoggerOptions;
	private_key_seed?: string;
	peer_discovery?: boolean;
	pubsub_peer_discovery_interval?: number;
}

type PeerDiscoveryFunction =
	| ((components: PubSubPeerDiscoveryComponents) => PeerDiscovery)
	| ((components: BootstrapComponents) => PeerDiscovery);

export class DRPNetworkNode {
	private _config?: DRPNetworkNodeConfig;
	private _node?: Libp2p;
	private _pubsub?: GossipSub;

	peerId = "";

	constructor(config?: DRPNetworkNodeConfig) {
		this._config = config;
		log = new Logger("drp::network", config?.log_config);
	}

	async start() {
		let privateKey = undefined;
		if (this._config?.private_key_seed) {
			const tmp = this._config.private_key_seed.padEnd(32, "0");
			privateKey = await generateKeyPairFromSeed("Ed25519", uint8ArrayFromString(tmp));
		}

		const _bootstrapNodesList = this._config?.bootstrap_peers
			? this._config.bootstrap_peers
			: BOOTSTRAP_NODES;

		const _peerDiscovery: Array<PeerDiscoveryFunction> = [
			pubsubPeerDiscovery({
				topics: ["drp::discovery"],
				interval: this._config?.pubsub_peer_discovery_interval || 1000,
			}),
		];

		const _bootstrapPeerID: string[] = [];
		if (_bootstrapNodesList.length) {
			_peerDiscovery.push(
				bootstrap({
					list: _bootstrapNodesList,
				})
			);
			for (const addr of _bootstrapNodesList) {
				const peerId = multiaddr(addr).getPeerId();
				if (!peerId) continue;
				_bootstrapPeerID.push(peerId);
			}
		}

		let _node_services: ServiceFactoryMap = {
			ping: ping(),
			dcutr: dcutr(),
			identify: identify(),
			identifyPush: identifyPush(),
			pubsub: gossipsub({
				doPX: true,
				allowPublishToZeroTopicPeers: true,
				scoreParams: createPeerScoreParams({
					IPColocationFactorWeight: 0,
					appSpecificScore: (peerId: string) => {
						if (_bootstrapNodesList.some((node) => node.includes(peerId))) {
							return 1000;
						}
						return 0;
					},
					topics: {
						"drp::discovery": createTopicScoreParams({
							topicWeight: 1,
						}),
					},
				}),
				fallbackToFloodsub: false,
			}),
		};

		if (this._config?.bootstrap) {
			_node_services = {
				..._node_services,
				autonat: autoNAT(),
				pubsub: gossipsub({
					// cf: https://github.com/libp2p/specs/blob/master/pubsub/gossipsub/gossipsub-v1.1.md#recommendations-for-network-operators
					D: 0,
					Dlo: 0,
					Dhi: 0,
					Dout: 0,
					doPX: true,
					ignoreDuplicatePublishError: true,
					allowPublishToZeroTopicPeers: true,
					scoreParams: createPeerScoreParams({
						topicScoreCap: 50,
						IPColocationFactorWeight: 0,
					}),
					fallbackToFloodsub: false,
				}),
				keychain: keychain(),
				// @ts-expect-error: it is not typed asdklasdjf;lasd
				autoTLS: autoTLS({
					autoConfirmAddress: true,
				}),
			};
		}

		const _bootstrap_services = {
			..._node_services,
			relay: circuitRelayServer({
				reservations: {
					maxReservations: Number.POSITIVE_INFINITY,
				},
			}),
		};

		this._node = await createLibp2p({
			privateKey,
			addresses: {
				listen: this._config?.listen_addresses
					? this._config.listen_addresses
					: ["/p2p-circuit", "/webrtc"],
				...(this._config?.announce_addresses ? { announce: this._config.announce_addresses } : {}),
			},
			connectionManager: {
				addressSorter: this._sortAddresses,
			},
			connectionEncrypters: [noise()],
			connectionGater: {
				denyDialMultiaddr: () => {
					return false;
				},
			},
			metrics: this._config?.browser_metrics ? devToolsMetrics() : undefined,
			peerDiscovery: _peerDiscovery,
			services: this._config?.bootstrap ? _bootstrap_services : _node_services,
			streamMuxers: [yamux()],
			transports: [
				circuitRelayTransport(),
				webRTC(),
				webRTCDirect(),
				webSockets({
					filter: filters.all,
				}),
				webTransport(),
			],
		});

		log.info("running on:", this._node.getMultiaddrs());
		if (!this._config?.bootstrap) {
			for (const addr of this._config?.bootstrap_peers || []) {
				try {
					await this._node.dial(multiaddr(addr));
				} catch (e) {
					log.error("::start::dial::error", e);
				}
			}
		}

		this._pubsub = this._node.services.pubsub as GossipSub;
		this.peerId = this._node.peerId.toString();

		log.info("::start: Successfuly started DRP network w/ peer_id", this.peerId);

		this._node.addEventListener("peer:connect", (e) =>
			log.info("::start::peer::connect", e.detail)
		);

		this._node.addEventListener("peer:discovery", (e) =>
			log.info("::start::peer::discovery", e.detail)
		);

		this._node.addEventListener("peer:identify", (e) =>
			log.info("::start::peer::identify", e.detail)
		);

		// needded as I've disabled the pubsubPeerDiscovery
		this._pubsub?.subscribe("drp::discovery");
		this._pubsub?.subscribe("drp::topic::discovery");
	}

	async stop() {
		await this._node?.stop();
	}

	async restart(config?: DRPNetworkNodeConfig) {
		await this.stop();
		if (config) this._config = config;
		await this.start();
	}

	async getPeerMultiaddrs(peerId: PeerId | string) {
		const peerIdObj: PeerId = typeof peerId === "string" ? peerIdFromString(peerId) : peerId;

		const peer = await this._node?.peerStore.get(peerIdObj);
		if (!peer) return [];
		return peer.addresses;
	}

	async isDialable(timeout = 5000) {
		return waitForEvent(async (resolve, reject) => {
			if (!this._node) {
				resolve(false);
				return;
			}

			if (await this._node?.isDialable(this._node.getMultiaddrs())) {
				resolve(true);
				return;
			}

			const checkDialable = async () => {
				try {
					if (await this._node?.isDialable(this._node.getMultiaddrs())) {
						resolve(true);
					}
				} catch (error) {
					reject(error);
				} finally {
					this._node?.removeEventListener("transport:listening", checkDialable);
				}
			};

			this._node.addEventListener("transport:listening", checkDialable);
		}, timeout);
	}

	async waitForPeer(peerId: string | PeerId, timeout = 5000) {
		const peerIdObj: PeerId = typeof peerId === "string" ? peerIdFromString(peerId) : peerId;

		return waitForEvent(async (resolve) => {
			if (this._node?.getPeers().some((p) => p.equals(peerIdObj))) {
				resolve(true);
				return;
			}

			const peerConnectListener = (e: CustomEvent<PeerId>) => {
				if (e.detail.equals(peerIdObj)) {
					this._node?.removeEventListener("peer:connect", peerConnectListener);
					resolve(true);
				}
			};
			this._node?.addEventListener("peer:connect", peerConnectListener);
		}, timeout);
	}

	async isSubscribed(topic: string, peerId: PeerId | string, timeout = 5000) {
		const peerIdObj: PeerId = typeof peerId === "string" ? peerIdFromString(peerId) : peerId;

		return waitForEvent(async (resolve) => {
			if (this._pubsub?.getSubscribers(topic).some((p) => p.equals(peerIdObj))) {
				resolve(true);
				return;
			}

			const subscriptionChangeListener = (e: CustomEvent<SubscriptionChangeData>) => {
				if (
					e.detail.subscriptions.some((s) => s.topic === topic && e.detail.peerId.equals(peerIdObj))
				) {
					resolve(true);
					this._pubsub?.removeEventListener("subscription-change", subscriptionChangeListener);
				}
			};

			this._pubsub?.addEventListener("subscription-change", subscriptionChangeListener);
		}, timeout);
	}

	private _sortAddresses(a: Address, b: Address) {
		const localRegex =
			/(^\/ip4\/127\.)|(^\/ip4\/10\.)|(^\/ip4\/172\.1[6-9]\.)|(^\/ip4\/172\.2[0-9]\.)|(^\/ip4\/172\.3[0-1]\.)|(^\/ip4\/192\.168\.)/;
		const aLocal = localRegex.test(a.toString());
		const bLocal = localRegex.test(b.toString());
		const aWebrtc = WebRTC.matches(a.multiaddr);
		const bWebrtc = WebRTC.matches(b.multiaddr);
		if (aLocal && !bLocal) return 1;
		if (!aLocal && bLocal) return -1;
		if (aWebrtc && !bWebrtc) return -1;
		if (!aWebrtc && bWebrtc) return 1;
		return 0;
	}

	changeTopicScoreParams(topic: string, params: TopicScoreParams) {
		if (!this._pubsub) return;
		this._pubsub.score.params.topics[topic] = params;
	}

	removeTopicScoreParams(topic: string) {
		if (!this._pubsub) return;
		delete this._pubsub.score.params.topics[topic];
	}

	subscribe(topic: string) {
		if (!this._node) {
			log.error("::subscribe: Node not initialized, please run .start()");
			return;
		}

		try {
			this._pubsub?.subscribe(topic);
			this._pubsub?.getPeers();
			log.info("::subscribe: Successfuly subscribed the topic", topic);
		} catch (e) {
			log.error("::subscribe:", e);
		}
	}

	unsubscribe(topic: string) {
		if (!this._node) {
			log.error("::unsubscribe: Node not initialized, please run .start()");
			return;
		}

		try {
			this._pubsub?.unsubscribe(topic);
			log.info("::unsubscribe: Successfuly unsubscribed the topic", topic);
		} catch (e) {
			log.error("::unsubscribe:", e);
		}
	}

	async connect(addr: MultiaddrInput | MultiaddrInput[]) {
		try {
			const addrs = (Array.isArray(addr) ? addr : [addr]).map(multiaddr);
			await this._node?.dial(addrs);
			log.info("::connect: Successfuly dialed", addr);
		} catch (e) {
			log.error("::connect:", e);
		}
	}

	async disconnect(peerId: string) {
		try {
			await this._node?.hangUp(multiaddr(`/p2p/${peerId}`));
			log.info("::disconnect: Successfuly disconnected", peerId);
		} catch (e) {
			log.error("::disconnect:", e);
		}
	}

	getBootstrapNodes() {
		return this._config?.bootstrap_peers ?? BOOTSTRAP_NODES;
	}

	getMultiaddrs() {
		return this._node?.getMultiaddrs().map((addr) => addr.toString());
	}

	getAllPeers() {
		const peers = this._node?.getPeers();
		if (!peers) return [];
		return peers.map((peer) => peer.toString());
	}

	getGroupPeers(group: string) {
		const peers = this._pubsub?.getSubscribers(group);
		if (!peers) return [];
		return peers.map((peer) => peer.toString());
	}

	async broadcastMessage(topic: string, message: Message) {
		try {
			const messageBuffer = Message.encode(message).finish();
			await this._pubsub?.publish(topic, messageBuffer);

			log.info("::broadcastMessage: Successfuly broadcasted message to topic", topic);
		} catch (e) {
			log.error("::broadcastMessage:", e);
		}
	}

	async sendMessage(peerId: string, message: Message) {
		try {
			const connection = await this._node?.dial([multiaddr(`/p2p/${peerId}`)]);
			const stream = <Stream>await connection?.newStream(DRP_MESSAGE_PROTOCOL);
			const messageBuffer = Message.encode(message).finish();
			await uint8ArrayToStream(stream, messageBuffer);
		} catch (e) {
			log.error("::sendMessage:", e);
		}
	}

	async sendGroupMessageRandomPeer(group: string, message: Message) {
		try {
			const peers = this._pubsub?.getSubscribers(group);
			if (!peers || peers.length === 0) throw Error("Topic wo/ peers");
			const peerId = peers[Math.floor(Math.random() * peers.length)];

			const connection = await this._node?.dial(peerId);
			const stream: Stream = (await connection?.newStream(DRP_MESSAGE_PROTOCOL)) as Stream;
			const messageBuffer = Message.encode(message).finish();
			await uint8ArrayToStream(stream, messageBuffer);
		} catch (e) {
			log.error("::sendMessageRandomTopicPeer:", e);
		}
	}

	addGroupMessageHandler(group: string, handler: EventCallback<CustomEvent<GossipsubMessage>>) {
		this._pubsub?.addEventListener("gossipsub:message", (e) => {
			if (group && e.detail.msg.topic !== group) return;
			handler(e);
		});
	}

	async addMessageHandler(handler: StreamHandler) {
		await this._node?.handle(DRP_MESSAGE_PROTOCOL, handler);
	}

	async addCustomMessageHandler(protocol: string | string[], handler: StreamHandler) {
		await this._node?.handle(protocol, handler);
	}
}
